#!/bin/bash

SCRIPT_PATH=/PATH_TO_SCRIPTS/
LOG_FILE=/var/log/cgroup_conf.log

MEMORY=`LANG=C free -b | grep "Mem:" | awk '{printf $2}'`
CPU_COUNT=`nproc`

if [ $# -eq 0 ]; then
  MODE=info
else
  MODE=$1
fi

function show_info() {
  local DEPTH=$1
  local GROUP=$2
  local NEXT_DEPTH=0

  if [ "x$DEPTH" = "x" ]; then
    DEPTH=0
  fi

  let NEXT_DEPTH=DEPTH+1

  for g in `(cd /sys/fs/cgroup/cpu/$GROUP; find * -maxdepth 0 -type d)`; do
    CPU_PERIOD=`cat /sys/fs/cgroup/cpu/$GROUP/$g/cpu.cfs_period_us`
    CPU_QUOTA=`cat /sys/fs/cgroup/cpu/$GROUP/$g/cpu.cfs_quota_us`
    CPU_SHARES=`cat /sys/fs/cgroup/cpu/$GROUP/$g/cpu.shares`

    let MAX_CPU_USAGE_PERC=100*CPU_QUOTA/CPU_PERIOD
    let MAX_CPU_USAGE_PERC=MAX_CPU_USAGE_PERC/CPU_COUNT

    MEMORY_USAGE_ABS=`cat /sys/fs/cgroup/memory/$GROUP/$g/memory.limit_in_bytes`
    let MEMORY_USAGE_PERC=100*MEMORY_USAGE_ABS
    let MEMORY_USAGE_PERC=MEMORY_USAGE_PERC/MEMORY

    TASKS=`cat /sys/fs/cgroup/cpu/$GROUP/$g/tasks | wc -l`

    for u in `seq 1 $DEPTH`; do
      echo -n "    "
    done

    echo "|---- $g $MAX_CPU_USAGE_PERC (Q/P:$CPU_QUOTA/$CPU_PERIOD) $CPU_SHARES $MEMORY_USAGE_PERC (L/M:$MEMORY_USAGE_ABS/$MEMORY) $TASKS $DEPTH"

    show_info $NEXT_DEPTH $GROUP/$g

  done
}

function show_list() {
  local GROUP=$1
  local SHOW_GROUP=$1

  for g in `(cd /sys/fs/cgroup/cpu/$GROUP; find * -maxdepth 0 -type d)`; do
    local SHOW_GROUP=`echo $GROUP/$g | sed -e "s#^/##g"`
    echo "$SHOW_GROUP"

    show_list $SHOW_GROUP
  done
}

function create_cgroup() {

  NEW_CGROUP=$1

  if [ ! -d /sys/fs/cgroup/cpu,cpuacct/$NEW_CGROUP ]; then
    #/usr/bin/cgm create cpu,cpuacct $NEW_CGROUP >> $LOG_FILE 2>&1
    cgcreate -g cpu,cpuacct:$NEW_CGROUP >> $LOG_FILE 2>&1
  fi

  if [ ! -d /sys/fs/cgroup/memory/$NEW_CGROUP ]; then
    #/usr/bin/cgm create memory $NEW_CGROUP >> $LOG_FILE 2>&1
    cgcreate -g memory:$NEW_CGROUP >> $LOG_FILE 2>&1
  fi
}

function configure_cgroup() {

  USER_CGROUP=$1

  echo "limit_user.sh: $USER_CGROUP" >> $LOG_FILE 2>&1

  ##############################################
  #                    CPU                     #
  ##############################################

  CPU_PERIOD=$2 # in ms
  if [ $CPU_PERIOD -eq 0 ]; then
    CPU_PERIOD=`cat /sys/fs/cgroup/cpu/cpu.cfs_period_us`
  fi

  cgset -r cpu.cfs_quota_us=-1 $USER_CGROUP >> $LOG_FILE 2>&1

  MAX_CPU_USAGE_PERC=$3 # in %

  echo "Set cpu_period=$CPU_PERIOD" >> $LOG_FILE 2>&1

  cgset -r cpu.cfs_period_us=$CPU_PERIOD $USER_CGROUP >> $LOG_FILE 2>&1

  CPU_PERIOD=`cat /sys/fs/cgroup/cpu/$USER_CGROUP/cpu.cfs_period_us`
  let CPU_QUOTA=CPU_PERIOD*CPU_COUNT*MAX_CPU_USAGE_PERC/100

  echo "Set cpu_quota=$CPU_QUOTA" >> $LOG_FILE 2>&1

  cgset -r cpu.cfs_quota_us=$CPU_QUOTA $USER_CGROUP >> $LOG_FILE 2>&1
  cgset -r cpuacct.usage=0 $USER_CGROUP >> $LOG_FILE 2>&1

  ##############################################
  #                  MEMORY                    #
  ##############################################

  if [ "x$4" != "x" ]; then
    MEMORY_USAGE_PERC=$4 # in %
    let MEMORY_USAGE_ABS=MEMORY*MEMORY_USAGE_PERC/100

    echo "Set memory_max=$MEMORY_USAGE_ABS" >> $LOG_FILE 2>&1

    cgset -r memory.limit_in_bytes=$MEMORY_USAGE_ABS $USER_CGROUP >> $LOG_FILE 2>&1
  fi

  if [ "x$5" != "x" ]; then
    CPU_SHARES=$5
    echo "Set cpu_shares=$CPU_SHARES" >> $LOG_FILE 2>&1
    cgset -r cpu.shares=$CPU_SHARES $USER_CGROUP >> $LOG_FILE 2>&1
  fi

  #
  # Using CGM
  #
  #/usr/bin/cgm setvalue cpu,cpuacct $USER_CGROUP/$USER cpu.cfs_period_us $CPU_PERIOD >> $LOG_FILE 2>&1
  #/usr/bin/cgm setvalue cpu,cpuacct $USER_CGROUP/$USER cpu.cfs_quota_us $CPU_QUOTA >> $LOG_FILE 2>&1
  #/usr/bin/cgm setvalue memory $USER_CGROUP/$USER memory.limit_in_bytes $MEMORY_USAGE_HARD >> $LOG_FILE 2>&1

}

function clear_cgroup() {
  local GROUP=$1
  local REF_CONTROLLER=$2

  if [ "x$GROUP" = "x/" ]; then
    GROUP=""
  fi
  if [ "x$REF_CONTROLLER" = "x" ]; then
    REF_CONTROLLER=cpu
  fi

  for g in `(cd /sys/fs/cgroup/$REF_CONTROLLER/$GROUP; find * -maxdepth 0 -type d)`; do
    clear_cgroup $GROUP/$g $REF_CONTROLLER
  done

  if [ "x$GROUP" != "x" ]; then
    for controller in cpu memory; do
      TASKS=`cat /sys/fs/cgroup/$controller/$GROUP/tasks | wc -l`
      SUBGROUPS=`(cd /sys/fs/cgroup/$controller/$GROUP; find * -maxdepth 0 -type d | wc -l)`

      if [ $TASKS -eq 0 ] && [ $SUBGROUPS -eq 0 ]; then
        # Empty so clear
        DELGROUP=`echo $GROUP | sed "s#^/##g"`
        echo "Delete $DELGROUP"
        cgdelete -g $controller:$DELGROUP > /dev/null 2>&1
      fi
    done
  fi
}

function move_tasks() {
  SRC_CGROUP=$1
  DST_CGROUP=$2

  for controller in memory cpu; do
    if [ -f /sys/fs/cgroup/$controller/$SRC_CGROUP/tasks ] && [ -f /sys/fs/cgroup/$controller/$DST_CGROUP/tasks ]; then
      for p in `cat /sys/fs/cgroup/$controller/$SRC_CGROUP/tasks`; do
        echo $p > /sys/fs/cgroup/$controller/$DST_CGROUP/tasks
      done
    fi
  done

}

case $MODE in
  "info")
    show_info
    ;;
  "list")
    show_list
    ;;
  "create")
     create_cgroup $2
     ;;
  "config")
     # TODO: use "shift 1" and all args
     configure_cgroup $2 $3 $4 $5 $6
     ;;
  "delete")
     delete_cgroup $2
     ;;
  "clear")
    clear_cgroup
    clear_cgroup "/" memory
    ;;
  "move")
     move_tasks $2 $3
     ;;
  "move_old")
     $0 create users
     $0 config users 100000 95 95 1024
     $0 create users/staff
     $0 config users/staff 100000 95 95 1024
     $0 create users/students
     $0 config users/students 100000 95 95 1024

     for old_g in `$0 list | grep -v users | awk -F/ '{ if ( NF == 2 ) { print \$0 }} '`; do
       echo $old_g
       $0 create users/$old_g
       $0 config users/$old_g 100000 90 90 1024

       $0 move $old_g users/$old_g
     done
     ;;
  "log")
     if [ "x$2" = "x" ]; then
       LEN=50
     else
       LEN=$2
     fi
     tail -n $LEN $LOG_FILE
     ;;
  *)
    echo "Use $0 [info|clear|set|create|move|delete]"
    echo "   config:  $0 group period_us cpu% mem% cpu-share (period_us=0 -> period_us=/sys/fs/cgroup/cpu/cpu.cfs_period_us"
    ;;
esac
